@@ -1,134 +1,65 @@
 import asyncio
+import logging
 import random
-import time
 from contextlib import suppress
 from operator import attrgetter
 from typing import Dict, List, Optional, Sequence, Set, Tuple
 
+import async_timeout
+
+from aioredis_cluster.cluster_state import (
+    ClusterState,
+    NodeClusterState,
+    _ClusterStateData,
+)
 from aioredis_cluster.command_info import (
     CommandsRegistry,
     create_registry,
     default_registry,
 )
-from aioredis_cluster.errors import (
-    ClusterStateError,
-    RedisError,
-    UncoveredSlotError,
-    network_errors,
-)
+from aioredis_cluster.errors import RedisError, network_errors
 from aioredis_cluster.log import logger
 from aioredis_cluster.pooler import Pooler
 from aioredis_cluster.structs import Address, ClusterNode, ClusterSlot
-from aioredis_cluster.typedef import SlotsResponse
+from aioredis_cluster.util import parse_info
 
 
-__all__ = [
-    "ClusterState",
+__all__ = (
     "ClusterManager",
     "slots_ranges",
     "create_cluster_state",
-]
+)
+
+
+SlotsResponse = List[List]
+CLUSTER_INFO_STATE_KEY = "cluster_state"
+CLUSTER_INFO_CURRENT_EPOCH_KEY = "cluster_current_epoch"
 
 
 def slots_ranges(slots: Sequence[ClusterSlot]) -> List[Tuple[int, int]]:
     return [(s.begin, s.end) for s in slots]
 
 
-class ClusterState:
-    __slots__ = (
-        "nodes",
-        "addrs",
-        "masters",
-        "replicas",
-        "slots",
-        "created_at",
-    )
-
-    def __init__(self) -> None:
-        self.nodes: Dict[Address, ClusterNode] = {}
-        self.addrs: List[Address] = []
-        self.masters: List[ClusterNode] = []
-        self.replicas: List[ClusterNode] = []
-        self.slots: List[ClusterSlot] = []
-        self.created_at = time.time()
-
-    def __repr__(self) -> str:
-        return "<{} {}>".format(type(self).__name__, self.repr_stats())
-
-    def repr_stats(self) -> str:
-        repr_parts = [
-            "created:{}".format(self.created_at),
-            "masters:{}".format(len(self.masters)),
-            "replicas:{}".format(len(self.replicas)),
-            "slot ranges:{}".format(len(self.slots)),
-        ]
-        return ", ".join(repr_parts)
-
-    def find_slot(self, slot: int) -> ClusterSlot:
-        # binary search
-        lo = 0
-        hi = len(self.slots)
-        while lo < hi:
-            mid = (lo + hi) // 2
-            cl_slot = self.slots[mid]
-
-            if cl_slot.end < slot:
-                lo = mid + 1
-            else:
-                hi = mid
-
-        if lo >= len(self.slots):
-            raise UncoveredSlotError(slot)
-
-        cl_slot = self.slots[lo]
-        if not cl_slot.in_range(slot):
-            raise UncoveredSlotError(slot)
-
-        return cl_slot
-
-    def slot_master(self, slot: int) -> ClusterNode:
-        return self.find_slot(slot).master
-
-    def slot_nodes(self, slot: int) -> List[ClusterNode]:
-        cl_slot = self.find_slot(slot)
-        nodes = [cl_slot.master]
-        nodes.extend(cl_slot.replicas)
-        return nodes
-
-    def random_slot_node(self, slot: int) -> ClusterNode:
-        return random.choice(self.slot_nodes(slot))
-
-    def random_slot_replica(self, slot: int) -> Optional[ClusterNode]:
-        replicas = self.find_slot(slot).replicas
-        if not replicas:
-            return None
-        return random.choice(replicas)
-
-    def random_master(self) -> ClusterNode:
-        if not self.masters:
-            raise ClusterStateError("no initialized masters")
-
-        return random.choice(self.masters)
-
-    def random_node(self) -> ClusterNode:
-        if not self.addrs:
-            raise ClusterStateError("no initialized nodes")
-
-        addr = random.choice(self.addrs)
-        return self.nodes[addr]
-
-
-def create_cluster_state(slots_resp: SlotsResponse) -> ClusterState:
-    state = ClusterState()
+def create_cluster_state(
+    slots_resp: SlotsResponse,
+    cluster_info: Dict[str, str],
+    state_from: Address,
+) -> ClusterState:
+    state_data = _ClusterStateData()
+    state_data.state = NodeClusterState(cluster_info[CLUSTER_INFO_STATE_KEY])
+    state_data.state_from = state_from
+    state_data.current_epoch = int(cluster_info[CLUSTER_INFO_CURRENT_EPOCH_KEY])
 
     nodes: Dict[Address, ClusterNode] = {}
+    master_replicas: Dict[Address, List[ClusterNode]] = {}
     masters_set: Set[ClusterNode] = set()
     replicas_set: Set[ClusterNode] = set()
 
     for slot_spec in slots_resp:
+        master_node: Optional[ClusterNode] = None
         slot_begin = int(slot_spec[0])
         slot_end = int(slot_spec[1])
-        slot_nodes: List[ClusterNode] = []
+        node_num = 0
         for node_num, node_spec in enumerate(slot_spec[2:], 1):
             node_addr = Address(node_spec[0], int(node_spec[1]))
             if node_addr not in nodes:
@@ -137,29 +68,44 @@ def create_cluster_state(slots_resp: SlotsResponse) -> ClusterState:
             else:
                 node = nodes[node_addr]
 
-            if node_num == 1 and node not in masters_set:
+            # first element is always master
+            if node_num == 1:
                 masters_set.add(node)
+                master_node = node
+                if master_node.addr not in master_replicas:
+                    master_replicas[master_node.addr] = []
             elif node_num > 1 and node not in replicas_set:
+                assert master_node is not None
                 replicas_set.add(node)
-            slot_nodes.append(node)
+                master_replicas[master_node.addr].append(node)
 
-        state.slots.append(
-            ClusterSlot(
-                begin=slot_begin, end=slot_end, master=slot_nodes[0], replicas=slot_nodes[1:]
+        if node_num == 0:
+            logger.error(
+                "CLUSTER SLOTS returns slot range %r without master node",
+                (slot_begin, slot_end),
+            )
+        else:
+            assert master_node is not None
+            state_data.slots.append(
+                ClusterSlot(
+                    begin=slot_begin,
+                    end=slot_end,
+                    master=master_node,
+                )
             )
-        )
 
-    state.slots.sort(key=attrgetter("begin", "end"))
-    state.nodes = nodes
-    state.masters = list(masters_set)
-    state.replicas = list(replicas_set)
-    state.addrs = list(nodes.keys())
+    state_data.slots.sort(key=attrgetter("begin", "end"))
+    state_data.nodes = nodes
+    state_data.masters = list(masters_set)
+    state_data.replicas = master_replicas
+    state_data.addrs = list(nodes.keys())
 
-    return state
+    return ClusterState(state_data)
 
 
 class ClusterManager:
     STATE_RELOAD_INTERVAL = 300.0
+    EXECUTE_TIMEOUT = 5.0
 
     def __init__(
         self,
@@ -168,6 +114,7 @@ class ClusterManager:
         *,
         state_reload_interval: float = None,
         follow_cluster: bool = None,
+        execute_timeout: float = None,
     ) -> None:
         self._startup_nodes = list(startup_nodes)
         self._pooler = pooler
@@ -180,6 +127,16 @@ class ClusterManager:
             follow_cluster = False
         self._follow_cluster = follow_cluster
 
+        if execute_timeout is None:
+            execute_timeout = self.EXECUTE_TIMEOUT
+        elif execute_timeout < 0:
+            raise ValueError("execute_timeout cannot be negative")
+
+        if execute_timeout > self._reload_interval:
+            execute_timeout = self._reload_interval
+
+        self._execute_timeout = execute_timeout
+
         self._reload_count = 0
         self._reload_event = asyncio.Event()
         loop = asyncio.get_event_loop()
@@ -209,31 +166,72 @@ class ClusterManager:
             with suppress(asyncio.CancelledError):
                 await self._reloader_task
 
-    async def _load_slots(self, addrs: Sequence[Address]) -> SlotsResponse:
-        first_err: Optional[Exception] = None
+    async def _fetch_state(self, addrs: Sequence[Address]) -> ClusterState:
+        if len(addrs) == 0:
+            raise RuntimeError("no addrs to fetch cluster state")
 
-        logger.debug("Trying to obtain cluster slots from addrs: %r", addrs)
+        last_err: Optional[BaseException] = None
+
+        if len(addrs) > 10:
+            # choose first minimum ten addrs
+            # addrs probable randomized
+            addrs = addrs[: max(10, len(addrs) // 2)]
+
+        logger.debug("Trying to obtain cluster state from addrs: %r", addrs)
 
         # get first successful cluster slots response
         for addr in addrs:
+            logger.info("Obtain cluster state from %s", addr)
             try:
                 pool = await self._pooler.ensure_pool(addr)
-                slots_resp = await pool.execute(b"CLUSTER", b"SLOTS", encoding="utf-8")
+                async with async_timeout.timeout(self._execute_timeout):
+                    # ensure one connection behaviour
+                    async with pool.get() as conn:
+                        raw_cluster_info: str = await conn.execute(
+                            b"CLUSTER", b"INFO", encoding="utf-8"
+                        )
+                        cluster_info = parse_info(raw_cluster_info)
+                        slots_resp = await conn.execute(b"CLUSTER", b"SLOTS", encoding="utf-8")
+
+            except asyncio.TimeoutError as e:
+                last_err = e
+                logger.warning("Getting cluster state from %s is timed out", addr)
+                continue
             except Exception as e:
-                if first_err is None:
-                    first_err = e
-                logger.warning("Unable to get cluster slots from %s: %r", addr, e)
+                last_err = e
+                logger.warning("Unable to get cluster state from %s: %r", addr, e)
                 continue
 
-            logger.debug("Cluster slots successful loaded from %s: %r", addr, slots_resp)
+            if cluster_info[CLUSTER_INFO_STATE_KEY] != NodeClusterState.OK.value:
+                logger.warning(
+                    'Node %s was return not "ok" cluster state "%s". Try next node',
+                    addr,
+                    cluster_info[CLUSTER_INFO_STATE_KEY],
+                )
+                continue
+
+            logger.debug("Cluster state successful loaded from %s: %r", addr, slots_resp)
 
             break
         else:
-            if first_err is not None:
-                logger.error("No available hosts to load cluster slots")
-                raise first_err
+            if last_err is not None:
+                logger.error("No available hosts to load cluster slots. Tried hosts: %r", addrs)
+                raise last_err
 
-        return slots_resp
+        state = create_cluster_state(slots_resp, cluster_info, addr)
+
+        if state.state is not NodeClusterState.OK:
+            logger.warning(
+                (
+                    "Cluster probably broken. Tried %d nodes and "
+                    'apply not "ok" (%s) cluster state from %s'
+                ),
+                len(addrs),
+                state.state.value,
+                addr,
+            )
+
+        return state
 
     async def _state_reloader(self) -> None:
         while True:
@@ -269,8 +267,8 @@ class ClusterManager:
             self._reload_event.clear()
 
     def _get_init_addrs(self, reload_id: int) -> List[Address]:
-        if self._follow_cluster and reload_id > 1 and self._state and self._state.addrs:
-            addrs = list(self._state.addrs)
+        if self._follow_cluster and reload_id > 1 and self._state and self._state._data.addrs:
+            addrs = list(self._state._data.addrs)
             random.shuffle(addrs)
         else:
             addrs = self._startup_nodes
@@ -280,18 +278,18 @@ class ClusterManager:
         commands: Optional[CommandsRegistry] = None
 
         init_addrs = self._get_init_addrs(reload_id)
-        slots_resp = await self._load_slots(init_addrs)
-        state = create_cluster_state(slots_resp)
+        state = await self._fetch_state(init_addrs)
 
         # initialize connections pool for every master node in new state
-        for node in state.masters:
+        for node in state._data.masters:
             await self._pooler.ensure_pool(node.addr)
 
         # choose random master node and load command specs from node
         pool = await self._pooler.ensure_pool(state.random_master().addr)
         # fetch commands only for first cluster state load
         if reload_id == 1:
-            raw_commands = await pool.execute(b"COMMAND", encoding="utf-8")
+            async with async_timeout.timeout(self._execute_timeout):
+                raw_commands = await pool.execute(b"COMMAND", encoding="utf-8")
             commands = create_registry(raw_commands)
             logger.debug("Found %d supported commands in cluster", commands.size())
 
@@ -299,19 +297,13 @@ class ClusterManager:
         self._state = state
         if commands is not None:
             self._commands = commands
-        logger.info(
-            "Loaded state: "
-            "%d slots ranges, "
-            "%d cluster nodes, "
-            "%d masters, "
-            "%d replicas "
-            "(reload_id=%d)",
-            len(state.slots),
-            len(state.nodes),
-            len(state.masters),
-            len(state.replicas),
-            reload_id,
-        )
+
+        if logger.isEnabledFor(logging.INFO):
+            logger.info(
+                "Loaded state: %s (reload_id=%d)",
+                state.repr_stats(),
+                reload_id,
+            )
 
         return self._state
 
