@@ -1,9 +1,15 @@
+import datetime
 from functools import lru_cache
 
 import pytest
 
 from aioredis_cluster.errors import ClusterStateError, UncoveredSlotError
-from aioredis_cluster.manager import ClusterState, create_cluster_state
+from aioredis_cluster.manager import (
+    ClusterState,
+    NodeClusterState,
+    _ClusterStateData,
+    create_cluster_state,
+)
 from aioredis_cluster.structs import Address
 
 from ._cluster_slots import SLOTS
@@ -11,7 +17,14 @@ from ._cluster_slots import SLOTS
 
 @lru_cache(None)
 def get_state():
-    return create_cluster_state(SLOTS)
+    return create_cluster_state(
+        SLOTS,
+        {
+            "cluster_state": "ok",
+            "cluster_current_epoch": "1",
+        },
+        Address("172.17.0.1", 6379),
+    )
 
 
 def get_nodes_addr(nodes):
@@ -31,7 +44,14 @@ def get_slots_ranges(slots):
 
 
 def test_create_cluster_state():
-    state = create_cluster_state(SLOTS)
+    state = create_cluster_state(
+        SLOTS,
+        {
+            "cluster_state": "ok",
+            "cluster_current_epoch": "1",
+        },
+        Address("172.17.0.1", 6379),
+    )
 
     addrs = sorted([Address("172.17.0.2", port) for port in range(7000, 7005)])
     masters_addrs = sorted(
@@ -59,21 +79,31 @@ def test_create_cluster_state():
         ]
     )
 
-    assert len(state.nodes) == 5
-    assert len(state.addrs) == 5
-    assert len(set(state.addrs)) == 5
-    assert len(state.masters) == 3
-    assert len(state.replicas) == 2
-    assert len(state.slots) == 7
-    assert sorted(state.addrs) == addrs
-    assert sorted(state.nodes.keys()) == addrs
-    assert get_nodes_addr(state.masters) == masters_addrs
-    assert get_nodes_addr(state.replicas) == replicas_addrs
-    assert get_slots_ranges(state.slots) == slot_ranges
+    state_data = state._data
+    assert len(state_data.nodes) == 5
+    assert len(state_data.addrs) == 5
+    assert len(set(state_data.addrs)) == 5
+    assert len(state_data.masters) == 3
+    assert len(state_data.replicas) == 3
+    assert sum(len(rs) for rs in state_data.replicas.values()) == 2
+    assert len(state_data.slots) == 7
+    assert sorted(state_data.addrs) == addrs
+    assert sorted(state_data.nodes.keys()) == addrs
+    assert get_nodes_addr(state_data.masters) == masters_addrs
+    assert get_nodes_addr([r for rs in state_data.replicas.values() for r in rs]) == replicas_addrs
+    assert get_slots_ranges(state_data.slots) == slot_ranges
+    assert state.state is NodeClusterState.OK
+    assert state.state_from == Address("172.17.0.1", 6379)
+    assert isinstance(state._data.created_at, datetime.datetime)
+    assert state._data.created_at_local > 0
+    assert state.current_epoch == 1
+
+    state.repr_stats()
+    str(state)
 
 
 def test_find_slot__empty_state():
-    state = ClusterState()
+    state = ClusterState(_ClusterStateData())
 
     with pytest.raises(UncoveredSlotError):
         state.find_slot(0)
@@ -106,7 +136,7 @@ def test_find_slot(slot, expect):
 
 
 def test_slot_master__empty_state():
-    state = ClusterState()
+    state = ClusterState(_ClusterStateData())
 
     with pytest.raises(UncoveredSlotError):
         state.slot_master(0)
@@ -136,7 +166,7 @@ def test_slot_master(slot, expect):
 
 
 def test_random_master__empty_state():
-    state = ClusterState()
+    state = ClusterState(_ClusterStateData())
 
     with pytest.raises(ClusterStateError):
         state.random_master()
@@ -148,11 +178,11 @@ def test_random_master(mocker):
     mocker.patch(state.__module__ + ".random.choice", side_effect=lambda s: s[1])
     result = state.random_master()
 
-    assert result is state.masters[1]
+    assert result is state.masters()[1]
 
 
 def test_random_node__empty_state():
-    state = ClusterState()
+    state = ClusterState(_ClusterStateData())
 
     with pytest.raises(ClusterStateError):
         state.random_node()
@@ -164,4 +194,29 @@ def test_random_node(mocker):
     mocker.patch(state.__module__ + ".random.choice", side_effect=lambda s: s[1])
     result = state.random_node()
 
-    assert result is state.nodes[state.addrs[1]]
+    assert result is state._data.nodes[state._data.addrs[1]]
+
+
+def test_state__with_fail_state():
+    state = create_cluster_state(
+        SLOTS,
+        {
+            "cluster_state": "fail",
+            "cluster_current_epoch": "1",
+        },
+        Address("172.17.0.1", 6379),
+    )
+
+    assert state.state is NodeClusterState.FAIL
+
+    state = create_cluster_state(
+        SLOTS,
+        {
+            "cluster_state": "foobar",
+            "cluster_current_epoch": "1",
+        },
+        Address("172.17.0.1", 6379),
+    )
+
+    assert state.state is NodeClusterState.UNKNOWN
+    assert state.current_epoch == 1
