@@ -1,28 +1,45 @@
 import asyncio
 import time
-from unittest import mock
 
-import asynctest
+import mock
 import pytest
 
+from aioredis_cluster.cluster_state import ClusterState, NodeClusterState
 from aioredis_cluster.manager import ClusterManager
 from aioredis_cluster.structs import Address, ClusterNode
 
+from ._cluster_slots import SLOTS
+
 
 @pytest.fixture
 def pooler_mock():
     def factory():
-        mocked_pool = mock.NonCallableMock()
-        mocked_pool.execute = asynctest.CoroutineMock()
+        conn = mock.NonCallableMock()
+        conn.execute = mock.AsyncMock(
+            return_value=object(),
+        )
+        pool = mock.NonCallableMock()
+        pool.execute = mock.AsyncMock(
+            return_value=object(),
+        )
+
+        acquirer_mock = mock.NonCallableMock()
+        acquirer_mock.__aenter__ = mock.AsyncMock(return_value=conn)
+        acquirer_mock.__aexit__ = mock.AsyncMock(return_value=None)
+        pool.get.return_value = acquirer_mock
 
         mocked = mock.NonCallableMock()
-        mocked.ensure_pool = asynctest.CoroutineMock(return_value=mocked_pool)
+        mocked.ensure_pool = mock.AsyncMock(return_value=pool)
+
+        mocked._pool = pool
+        mocked._conn = conn
+
         return mocked
 
     return factory
 
 
-async def test_require_reload_state(pooler_mock, loop):
+async def test_require_reload_state(pooler_mock):
     pooler = pooler_mock()
     manager = ClusterManager(["addr1", "addr2"], pooler)
 
@@ -35,15 +52,15 @@ async def test_require_reload_state(pooler_mock, loop):
     await manager.close()
 
 
-async def test_state_reloader(mocker, pooler_mock, loop):
+async def test_state_reloader(mocker, pooler_mock, event_loop):
     pooler = pooler_mock()
     manager = ClusterManager(["addr1", "addr2"], pooler)
 
-    load_state_fut = loop.create_future()
+    load_state_fut = event_loop.create_future()
     mocked_load_state = mocker.patch.object(
         manager,
         "_load_state",
-        new=asynctest.CoroutineMock(side_effect=lambda reload_id: load_state_fut.set_result(None)),
+        new=mock.AsyncMock(side_effect=lambda reload_id: load_state_fut.set_result(None)),
     )
 
     manager.require_reload_state()
@@ -59,7 +76,7 @@ async def test_get_state__no_state_load_state(mocker, pooler_mock):
     pooler = pooler_mock()
     manager = ClusterManager(["addr1", "addr2"], pooler)
 
-    mocked_load_state = mocker.patch.object(manager, "_load_state", new=asynctest.CoroutineMock())
+    mocked_load_state = mocker.patch.object(manager, "_load_state", new=mock.AsyncMock())
 
     state = await manager.get_state()
 
@@ -74,7 +91,7 @@ async def test_get_state__state_exists(mocker, pooler_mock):
     manager = ClusterManager(["addr1", "addr2"], pooler)
 
     state = mock.NonCallableMock()
-    state.created_at = time.time()
+    state._data.created_at = time.time()
     manager._state = state
 
     mocked_require_reload_state = mocker.patch.object(manager, "require_reload_state")
@@ -107,24 +124,29 @@ async def test_get_state__state_exists_and_require_reload_state(mocker, pooler_m
     await manager.close()
 
 
-async def test_load_state(mocker, pooler_mock, loop):
+async def test_load_state(mocker, pooler_mock):
     pooler = pooler_mock()
 
     mocker.patch(ClusterManager.__module__ + ".random.sample", return_value=["addr1", "addr2"])
 
     manager = ClusterManager(["addr1", "addr2"], pooler)
 
-    mocked_load_slots = mocker.patch.object(manager, "_load_slots", new=asynctest.CoroutineMock())
-    mocked_create_cluster_state = mocker.patch(manager.__module__ + ".create_cluster_state")
-    mocked_state = mocked_create_cluster_state.return_value
-    mocked_state.masters = [
+    mocked_state = mock.NonCallableMock()
+    mocked_state._data.masters = [
         ClusterNode(Address("master1", 6666), "master1_id"),
         ClusterNode(Address("master2", 7777), "master2_id"),
     ]
-    mocked_state.addrs = [object(), object()]
+    mocked_state._data.addrs = [object(), object()]
     mocked_state.random_master.return_value = ClusterNode(
         Address("random_master", 5555), "random_master_id"
     )
+    mocked_fetch_state = mocker.patch.object(
+        manager,
+        "_fetch_state",
+        new=mock.AsyncMock(
+            return_value=mocked_state,
+        ),
+    )
 
     mocked_create_registry = mocker.patch(manager.__module__ + ".create_registry")
 
@@ -133,7 +155,7 @@ async def test_load_state(mocker, pooler_mock, loop):
     assert result is mocked_state
     assert result is manager._state
     assert manager._commands is mocked_create_registry.return_value
-    mocked_load_slots.assert_called_once_with(["addr1", "addr2"])
+    mocked_fetch_state.assert_called_once_with(["addr1", "addr2"])
     mocked_create_registry.assert_called_once_with(
         pooler.ensure_pool.return_value.execute.return_value
     )
@@ -150,7 +172,7 @@ async def test_init(mocker, pooler_mock):
     pooler = pooler_mock()
     manager = ClusterManager(["addr1", "addr2"], pooler)
 
-    mocked_load_state = mocker.patch.object(manager, "_load_state", new=asynctest.CoroutineMock())
+    mocked_load_state = mocker.patch.object(manager, "_load_state", new=mock.AsyncMock())
 
     await manager._init()
 
@@ -160,21 +182,46 @@ async def test_init(mocker, pooler_mock):
     await manager.close()
 
 
-async def test_load_slots__first_response(mocker, pooler_mock):
+async def test_fetch_state__first_response(pooler_mock, mocker):
     pooler = pooler_mock()
-    pool = pooler.ensure_pool.return_value
+    pool = pooler._pool
+    conn = pooler._conn
+    cluster_slots_resp = object()
+    conn.execute.side_effect = [
+        "cluster_state:ok\ncluster_current_epoch:1",
+        cluster_slots_resp,
+    ]
     manager = ClusterManager(["addr1"], pooler)
+    mocked_create_cluster_state = mocker.patch(ClusterManager.__module__ + ".create_cluster_state")
 
-    result = await manager._load_slots(["addr2", "addr3"])
+    result = await manager._fetch_state(["addr2", "addr3"])
 
-    assert result is pool.execute.return_value
+    assert result is mocked_create_cluster_state.return_value
     pooler.ensure_pool.assert_called_once_with("addr2")
-    pool.execute.assert_called_once_with(b"CLUSTER", b"SLOTS", encoding="utf-8")
+    pool.get.asssert_called_once()
+    assert conn.execute.await_count == 2
+    execute_calls = conn.execute.call_args_list
+    assert execute_calls[0] == mock.call(b"CLUSTER", b"INFO", encoding="utf-8")
+    assert execute_calls[1] == mock.call(b"CLUSTER", b"SLOTS", encoding="utf-8")
 
 
-async def test_load_slots__with_error_but_success(mocker, pooler_mock):
+async def test_fetch_state__with_fail_state(pooler_mock, mocker):
     pooler = pooler_mock()
-    pool = pooler.ensure_pool.return_value
+    conn = pooler._conn
+    conn.execute.side_effect = [
+        "cluster_state:fail\ncluster_current_epoch:1",
+        SLOTS,
+    ]
+    manager = ClusterManager(["addr1"], pooler)
+    result = await manager._fetch_state(["addr2"])
+
+    assert result.state is NodeClusterState.FAIL
+
+
+async def test_fetch_state__with_error_but_success(pooler_mock):
+    pooler = pooler_mock()
+    conn = pooler._conn
+    pool = pooler._pool
 
     def ensure_pool_se(addr):
         if addr == "addr1":
@@ -183,47 +230,68 @@ async def test_load_slots__with_error_but_success(mocker, pooler_mock):
 
     pooler.ensure_pool.side_effect = ensure_pool_se
 
-    def pool_execute_se(*args, **kwargs):
-        if pool.execute.call_count == 1:
-            raise RuntimeError("execute")
-        return pool.execute.return_value
-
-    pool.execute.side_effect = pool_execute_se
+    conn.execute.side_effect = [
+        # first try
+        "cluster_state:fail\ncluster_current_epoch:1",
+        object(),
+        # second try
+        "cluster_state:ok\ncluster_current_epoch:1",
+        RuntimeError("execute fail"),
+        # success third try
+        "cluster_state:ok\ncluster_current_epoch:1",
+        SLOTS,
+    ]
 
     manager = ClusterManager(["addr1"], pooler)
 
-    result = await manager._load_slots(["addr1", "addr2", "addr3"])
+    result = await manager._fetch_state(["addr1", "addr2", "addr3", "addr4"])
 
-    assert result is pool.execute.return_value
-    assert pooler.ensure_pool.call_count == 3
-    assert pool.execute.call_count == 2
+    assert isinstance(result, ClusterState)
+    assert result.state is NodeClusterState.OK
+    assert result.state_from == "addr4"
+    assert pooler.ensure_pool.await_count == 4
+    assert conn.execute.await_count == 6
     pooler.ensure_pool.assert_has_calls(
         [
             mock.call("addr1"),
             mock.call("addr2"),
             mock.call("addr3"),
+            mock.call("addr4"),
         ]
     )
-    pool.execute.assert_has_calls(
+    conn.execute.assert_has_calls(
         [
-            mock.call(b"CLUSTER", b"SLOTS", encoding="utf-8"),
+            mock.call(b"CLUSTER", b"INFO", encoding="utf-8"),
             mock.call(b"CLUSTER", b"SLOTS", encoding="utf-8"),
         ]
     )
 
 
-async def test_load_slots__with_error(mocker, pooler_mock):
+async def test_fetch_state__with_error(pooler_mock):
     pooler = pooler_mock()
+    pool = pooler._pool
+    conn = pooler._conn
 
-    def ensure_pool_se(addr):
-        raise RuntimeError(addr)
+    pooler.ensure_pool.side_effect = [
+        RuntimeError("first error"),
+        pool,
+        pool,
+        pool,
+    ]
 
-    pooler.ensure_pool.side_effect = ensure_pool_se
+    pool.get.return_value.__aenter__.side_effect = [
+        RuntimeError("second error"),
+        conn,
+        conn,
+    ]
+
+    execute_resp = mock.NonCallableMagicMock()
+    conn.execute.side_effect = [RuntimeError("third error"), execute_resp]
 
     manager = ClusterManager(["addr1"], pooler)
 
-    with pytest.raises(RuntimeError, match="addr1"):
-        await manager._load_slots(["addr1", "addr2", "addr3"])
+    with pytest.raises(RuntimeError, match="third error"):
+        await manager._fetch_state(["addr1", "addr2", "addr3"])
 
     assert pooler.ensure_pool.call_count == 3
     pooler.ensure_pool.assert_has_calls(
@@ -264,7 +332,8 @@ async def test_get_init_addrs__cluster_following(
     manager = ClusterManager(["startup1", "startup2"], pooler)
     manager._follow_cluster = follow_cluster
     if state_addrs is not None:
-        manager._state = mock.NonCallableMock(addrs=state_addrs)
+        manager._state = mock.NonCallableMock()
+        manager._state._data.addrs = state_addrs
 
     result = manager._get_init_addrs(reload_id)
 
@@ -277,7 +346,7 @@ async def test_reload_state(mocker, pooler_mock):
 
     state = object()
     mocked_load_state = mocker.patch.object(
-        manager, "_load_state", new=asynctest.CoroutineMock(return_value=state)
+        manager, "_load_state", new=mock.AsyncMock(return_value=state)
     )
 
     result = await manager.reload_state()
